@@ -1,6 +1,6 @@
 /***********************************************************************
 Sandbox - Vrui application to drive an augmented reality sandbox.
-Copyright (c) 2012-2015 Oliver Kreylos
+Copyright (c) 2012-2016 Oliver Kreylos
 
 This file is part of the Augmented Reality Sandbox (SARndbox).
 
@@ -23,8 +23,8 @@ Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 #define SANDBOX_INCLUDED
 
 #include <Threads/TripleBuffer.h>
-#include <USB/Context.h>
 #include <Geometry/Box.h>
+#include <Geometry/OrthonormalTransformation.h>
 #include <Geometry/ProjectiveTransformation.h>
 #include <GL/gl.h>
 #include <GL/GLColorMap.h>
@@ -40,7 +40,7 @@ Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 #include <Kinect/FrameBuffer.h>
 #include <Kinect/FrameSource.h>
 
-#include "RainMaker.h"
+#include "Types.h"
 
 /* Forward declarations: */
 namespace Misc {
@@ -60,78 +60,28 @@ namespace Kinect {
 class Camera;
 }
 class FrameFilter;
+class DepthImageRenderer;
+class ElevationColorMap;
+class DEM;
 class SurfaceRenderer;
 class WaterTable2;
+class HandExtractor;
 typedef Misc::FunctionCall<GLContextData&> AddWaterFunction;
+class WaterRenderer;
 
 class Sandbox:public Vrui::Application,public GLObject
 	{
 	/* Embedded classes: */
 	private:
-	typedef Geometry::Box<double,3> Box; // Type for bounding boxes
-	typedef Geometry::ProjectiveTransformation<double,3> PTransform; // Type for projective transformations
-	
-	class WaterTool;
-	typedef Vrui::GenericToolFactory<WaterTool> WaterToolFactory;
-	
-	class WaterTool:public Vrui::Tool,public Vrui::Application::Tool<Sandbox>
-		{
-		friend class Vrui::GenericToolFactory<WaterTool>;
-		
-		/* Elements: */
-		private:
-		static WaterToolFactory* factory; // Pointer to the factory object for this class
-		
-		/* Constructors and destructors: */
-		public:
-		static WaterToolFactory* initClass(Vrui::ToolManager& toolManager);
-		WaterTool(const Vrui::ToolFactory* factory,const Vrui::ToolInputAssignment& inputAssignment);
-		virtual ~WaterTool(void);
-		
-		/* Methods from class Vrui::Tool: */
-		virtual const Vrui::ToolFactory* getFactory(void) const;
-		virtual void buttonCallback(int buttonSlotIndex,Vrui::InputDevice::ButtonCallbackData* cbData);
-		};
-	
-	class LocalWaterTool;
-	typedef Vrui::GenericToolFactory<LocalWaterTool> LocalWaterToolFactory;
-	
-	class LocalWaterTool:public Vrui::Tool,public Vrui::Application::Tool<Sandbox>,public Vrui::TransparentObject
-		{
-		friend class Vrui::GenericToolFactory<LocalWaterTool>;
-		
-		/* Elements: */
-		private:
-		static LocalWaterToolFactory* factory; // Pointer to the factory object for this class
-		
-		const AddWaterFunction* addWaterFunction; // Render function registered with the water table
-		GLfloat adding; // Amount of data added or removed from the water table
-		
-		/* Constructors and destructors: */
-		public:
-		static LocalWaterToolFactory* initClass(Vrui::ToolManager& toolManager);
-		LocalWaterTool(const Vrui::ToolFactory* factory,const Vrui::ToolInputAssignment& inputAssignment);
-		virtual ~LocalWaterTool(void);
-		
-		/* Methods from class Vrui::Tool: */
-		virtual void initialize(void);
-		virtual void deinitialize(void);
-		virtual const Vrui::ToolFactory* getFactory(void) const;
-		virtual void buttonCallback(int buttonSlotIndex,Vrui::InputDevice::ButtonCallbackData* cbData);
-		
-		/* Methods from class Vrui::TransparentObject: */
-		virtual void glRenderActionTransparent(GLContextData& contextData) const;
-		
-		/* New methods: */
-		void addWater(GLContextData& contextData) const; // Function to render geometry that adds water to the water table
-		};
+	typedef Geometry::Box<Scalar,3> Box; // Type for bounding boxes
+	typedef Geometry::OrthonormalTransformation<Scalar,3> ONTransform; // Type for rigid body transformations
+	typedef Kinect::FrameSource::DepthCorrection::PixelCorrection PixelDepthCorrection; // Type for per-pixel depth correction factors
 	
 	struct DataItem:public GLObject::DataItem
 		{
 		/* Elements: */
 		public:
-		GLuint heightColorMapObject; // Color map for the height map shader
-		unsigned int heightColorMapVersion; // Version number of height color map in texture object
+		double waterTableTime; // Simulation time stamp of the water table in this OpenGL context
 		GLsizei shadowBufferSize[2]; // Size of the shadow rendering frame buffer
 		GLuint shadowFramebufferObject; // Frame buffer object to render shadow maps
 		GLuint shadowDepthTextureObject; // Depth texture for the shadow rendering frame buffer
@@ -141,39 +91,63 @@ class Sandbox:public Vrui::Application,public GLObject
 		virtual ~DataItem(void);
 		};
 	
-	friend class WaterTool;
+	struct RenderSettings // Structure to hold per-window rendering settings
+		{
+		/* Elements: */
+		public:
+		bool fixProjectorView; // Flag whether to allow viewpoint navigation or always render from the projector's point of view
+		PTransform projectorTransform; // The calibrated projector transformation matrix for fixed-projection rendering
+		bool projectorTransformValid; // Flag whether the projector transformation is valid
+		bool hillshade; // Flag whether to use augmented reality hill shading
+		GLMaterial surfaceMaterial; // Material properties to render the surface in hill shading mode
+		bool useShadows; // Flag whether to use shadows in augmented reality hill shading
+		ElevationColorMap* elevationColorMap; // Pointer to an elevation color map
+		bool useContourLines; // Flag whether to draw elevation contour lines
+		GLfloat contourLineSpacing; // Spacing between adjacent contour lines in cm
+		bool renderWaterSurface; // Flag whether to render the water surface as a geometric surface
+		GLfloat waterOpacity; // Opacity factor for water when rendered as texture
+		SurfaceRenderer* surfaceRenderer; // Surface rendering object for this window
+		WaterRenderer* waterRenderer; // A renderer to render the water surface as geometry
+		
+		/* Constructors and destructors: */
+		RenderSettings(void); // Creates default rendering settings
+		RenderSettings(const RenderSettings& source); // Copy constructor
+		~RenderSettings(void); // Destroys rendering settings
+		
+		/* Methods: */
+		void loadProjectorTransform(const char* projectorTransformName); // Loads a projector transformation from the given file
+		void loadHeightMap(const char* heightMapName); // Loads the selected height map
+		};
+	
+	friend class GlobalWaterTool;
 	friend class LocalWaterTool;
+	friend class DEMTool;
 	
 	/* Elements: */
 	private:
-	USB::Context usbContext; // USB context for the Kinect camera device
-	Kinect::Camera* camera; // The Kinect camera device
+	Kinect::FrameSource* camera; // The Kinect camera device
 	unsigned int frameSize[2]; // Width and height of the camera's depth frames
+	PixelDepthCorrection* pixelDepthCorrection; // Buffer of per-pixel depth correction coefficients
 	Kinect::FrameSource::IntrinsicParameters cameraIps; // Intrinsic parameters of the Kinect camera
 	FrameFilter* frameFilter; // Processing object to filter raw depth frames from the Kinect camera
 	bool pauseUpdates; // Pauses updates of the topography
 	Threads::TripleBuffer<Kinect::FrameBuffer> filteredFrames; // Triple buffer for incoming filtered depth frames
-	PTransform projectorTransform; // The calibrated projector transformation matrix
-	GLMaterial surfaceMaterial; // Material properties to render the surface
-	GLColorMap heightMap; // The height color map
-	unsigned int heightMapVersion; // Version number of height map
+	DepthImageRenderer* depthImageRenderer; // Object managing the current filtered depth image
+	ONTransform boxTransform; // Transformation from camera space to baseplane space (x along long sandbox axis, z up)
 	Box bbox; // Bounding box around the surface
-	SurfaceRenderer* surfaceRenderer; // Renderer for the surface
 	WaterTable2* waterTable; // Water flow simulation object
 	double waterSpeed; // Relative speed of water flow simulation
-	int waterMaxSteps; // Maximum number of water simulation steps per frame
+	unsigned int waterMaxSteps; // Maximum number of water simulation steps per frame
 	GLfloat rainStrength; // Amount of water deposited by rain tools and objects on each water simulation step
-	FrameFilter* rmFrameFilter; // Second frame filter for the rain maker object
-	RainMaker* rainMaker; // Object to detect objects floating above the sandbox to make rain
-	Threads::TripleBuffer<RainMaker::BlobList> rainObjects; // Triple buffer of detected rain objects
+	HandExtractor* handExtractor; // Object to detect splayed hands above the sand surface to make rain
 	const AddWaterFunction* addWaterFunction; // Render function registered with the water table
 	bool addWaterFunctionRegistered; // Flag if the water adding function is currently registered with the water table
-	bool fixProjectorView; // Flag whether to allow viewpoint navigation or always render from the projector's point of view
-	bool hillshade; // Flag whether to use augmented reality hill shading
-	bool useShadows; // Flag whether to use shadows in augmented reality hill shading
-	bool useHeightMap; // Flag whether to color map the surface by elevation
-	SurfaceRenderer* waterRenderer; // A second surface renderer to render the water surface directly
+	std::vector<RenderSettings> renderSettings; // List of per-window rendering settings
 	Vrui::Lightsource* sun; // An external fixed light source
+	Vrui::Point navCenter;
+	Vrui::Scalar navSize;
+	Vrui::Vector navUp;
+	DEM* activeDem; // The currently active DEM
 	GLMotif::PopupMenu* mainMenu;
 	GLMotif::ToggleButton* pauseUpdatesToggle;
 	GLMotif::PopupWindow* waterControlDialog;
@@ -186,7 +160,7 @@ class Sandbox:public Vrui::Application,public GLObject
 	/* Private methods: */
 	void rawDepthFrameDispatcher(const Kinect::FrameBuffer& frameBuffer); // Callback receiving raw depth frames from the Kinect camera; forwards them to the frame filter and rain maker objects
 	void receiveFilteredFrame(const Kinect::FrameBuffer& frameBuffer); // Callback receiving filtered depth frames from the filter object
-	void receiveRainObjects(const RainMaker::BlobList& newRainObjects); // Callback receiving extracted rain objects from the rain maker
+	void toggleDEM(DEM* dem); // Sets or toggles the currently active DEM
 	void addWater(GLContextData& contextData) const; // Function to render geometry that adds water to the water table
 	void pauseUpdatesCallback(GLMotif::ToggleButton::ValueChangedCallbackData* cbData);
 	void showWaterControlDialogCallback(Misc::CallbackData* cbData);
@@ -195,7 +169,6 @@ class Sandbox:public Vrui::Application,public GLObject
 	void waterAttenuationSliderCallback(GLMotif::TextFieldSlider::ValueChangedCallbackData* cbData);
 	GLMotif::PopupMenu* createMainMenu(void);
 	GLMotif::PopupWindow* createWaterControlDialog(void);
-	bool loadHeightColorMap(const char* heightColorMapFileName); // Loads a new height color map from a file of the given name
 	
 	/* Constructors and destructors: */
 	public:
@@ -203,8 +176,10 @@ class Sandbox:public Vrui::Application,public GLObject
 	virtual ~Sandbox(void);
 	
 	/* Methods from Vrui::Application: */
+	virtual void toolDestructionCallback(Vrui::ToolManager::ToolDestructionCallbackData* cbData);
 	virtual void frame(void);
 	virtual void display(GLContextData& contextData) const;
+	virtual void resetNavigation(void);
 	virtual void eventCallback(EventID eventId,Vrui::InputDevice::ButtonCallbackData* cbData);
 	
 	/* Methods from GLObject: */
